import 'dart:async';

import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

typedef Future QueryFunction();

class Query {
  String key;
  QueryFunction fn;
  StreamController _dataStream = StreamController();

  Query(this.key, this.fn) {
    invalidate();
  }

  void watch(Function watcher) {
    _dataStream.stream.listen((data) => watcher(data));
  }

  void invalidate() {
    fn().then((data) => _dataStream.add(data));
  }
}

class QueryCache {
  List<Query> _queries = [];

  List<Query> getQueriesByKey(String key) {
    return _queries.where((q) => q.key == key).toList();
  }

  void add(Query query) {
    _queries.add(query);
  }

  void invalidate(String key) {
    getQueriesByKey(key).forEach((q) => q.invalidate());
  }
}

class QueryCacheProvider extends StatefulWidget {
  final Widget child;
  QueryCacheProvider({this.child});

  @override
  State<StatefulWidget> createState() => QueryCacheProviderState();
}

class QueryCacheProviderState extends State<QueryCacheProvider> {
  QueryCache _queryCache;

  @override
  void initState() {
    _queryCache = QueryCache();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Provider.value(
      value: _queryCache,
      child: widget.child,
    );
  }
}
